-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("tagQuestLog","AceEvent-3.0", "AceHook-3.0")

-----------------------
--- Library loading ---
-----------------------
local LTQL = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-TagQuestLog")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local min, sub, len, match, tonumber = math.min, string.sub, string.len, string.match, tonumber
local GetNumQuestLogEntries, HybridScrollFrame_GetOffset = GetNumQuestLogEntries, HybridScrollFrame_GetOffset

-----------------
--- variables ---
-----------------
module.modName = LTQL["Quest Log Tag"]

function module:OnInitialize()
	self:SetEnabledState(core:getModuleEnabled("tagQuestLog"))
	core.moduleState["tagQuestLog"] = core:getModuleEnabled("tagQuestLog")
end

function module:OnEnable()
	self:SecureHook("QuestLog_Update")
	QuestLogScrollFrame:HookScript("OnMouseWheel", function() module:QuestLog_Update() end)
	QuestLogScrollFrameScrollBar:HookScript("OnValueChanged", function() module:QuestLog_Update() end)
end

function module:QuestLog_Update()
	if not QuestLogFrame:IsShown() then return end
	local buttons = QuestLogScrollFrame.buttons
	local numButtons = #buttons
	local offset, numEntries, numQuests = HybridScrollFrame_GetOffset(QuestLogScrollFrame), GetNumQuestLogEntries()
	local zone
	for i=1, numButtons do
		local index = i + offset
		if index > numEntries then break end
		local questLogTitle = buttons[i]
		local questTitleTag, questCheck, questNumGroupMates = questLogTitle.tag, questLogTitle.check, questLogTitle.groupMates
		if questCheck then questCheck:Hide() end
		local uid = tonumber(match(GetQuestLink(index) or '', 'quest:(%d+)'))
		if uid and core.quests[uid] then
			questLogTitle:SetText("  "..core.quests[uid].tag..core.quests[uid].title)
			if core.quests[uid].playerComplete == 0 then
				questTitleTag:Hide()
			end
			if IsQuestWatched(index) then
				questCheck:Show()
			end
			local numMates = questNumGroupMates:GetText() or ""
			local numMatesLength = len(numMates)
			if numMatesLength > 0 then
				numMates = sub(numMates, 2, numMatesLength - 1)
				questNumGroupMates:SetText("<"..numMates..">")
				questNumGroupMates:Show()
			else
				questNumGroupMates:Hide()
			end
			QuestLogTitleButton_Resize(questLogTitle)
		end
	end
end